# generated by datamodel-codegen:
#   filename:  liquid-chromatography.schema.json
#   timestamp: 2024-12-09T23:59:03+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCentimeter,
    TQuantityValueCounts,
    TQuantityValueCountsTimesSecond,
    TQuantityValueCubicMillimeter,
    TQuantityValueHertz,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliAbsorbanceUnitTimesMilliliter,
    TQuantityValueMilliAbsorbanceUnitTimesSecond,
    TQuantityValueMilliliter,
    TQuantityValueMillimeter,
    TQuantityValueMillivolt,
    TQuantityValueMillivoltTimesSecond,
    TQuantityValueNanoCoulomb,
    TQuantityValueNanoCoulombTimesSecond,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePercent,
    TQuantityValuePicoAmpere,
    TQuantityValuePicoAmpereTimesSecond,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TClass,
    TDatacube,
    TDatacubeData,
    TDatacubeStructure,
    TDateTimeStampValue,
    TDateTimeValue,
    TQuantityValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None


@dataclass(kw_only=True)
class AdmCoreREC202309ManifestSchema:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class StatisticsDocumentItem:
    statistical_feature: TClass


@dataclass(kw_only=True)
class StatisticsAggregateDocument:
    statistics_document: list[StatisticsDocumentItem] | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    UNC_path: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None
    pump_model_number: TStringValue | None = None
    detector_model_number: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    detection_type: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    model_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None
    detector_offset_setting: TQuantityValue | None = None
    detector_sampling_rate_setting: TQuantityValueHertz | None = None
    detector_wavelength_setting: TQuantityValueNanometer | None = None
    detector_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_wavelength_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_reference_bandwidth_setting: TQuantityValueNanometer | None = (
        None
    )
    electronic_absorbance_reference_wavelength_setting: TQuantityValueNanometer | None = (
        None
    )
    solvent_concentration_data_cube: SolventConcentrationDataCube | None = None
    pre_column_pressure_data_cube: PreColumnPressureDataCube | None = None
    post_column_pressure_data_cube: PostColumnPressureDataCube | None = None
    sample_pressure_data_cube: SamplePressureDataCube | None = None
    system_pressure_data_cube: SystemPressureDataCube | None = None
    excitation_wavelength_setting: TQuantityValueNanometer | None = None
    excitation_bandwidth_setting: TQuantityValueNanometer | None = None
    total_measurement_time_setting: TQuantityValueSecondTime | None = None
    read_interval_setting: TQuantityValueSecondTime | None = None
    number_of_scans_setting: TQuantityValueNumber | None = None
    system_flow_rate_data_cube: SystemFlowRateDataCube | None = None
    sample_flow_rate_data_cube: SampleFlowRateDataCube | None = None
    temperature_profile_data_cube: TemperatureProfileDataCube | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    description: Any | None = None
    batch_identifier: TStringValue | None = None
    sample_role_type: TClass | None = None
    written_name: TStringValue | None = None


@dataclass(kw_only=True)
class Peak:
    peak_end: TQuantityValueMilliliter | TQuantityValueSecondTime | None = None
    identifier: TStringValue | None = None
    relative_peak_height: TQuantityValuePercent | None = None
    written_name: TStringValue | None = None
    peak_height: TQuantityValue | TQuantityValueCounts | TQuantityValueMilliAbsorbanceUnit | TQuantityValueMillivolt | TQuantityValueNanoCoulomb | TQuantityValuePicoAmpere | None = (
        None
    )
    capacity_factor__chromatography_: TQuantityValueUnitless | None = None
    peak_area: TQuantityValue | TQuantityValueCountsTimesSecond | TQuantityValueMilliAbsorbanceUnitTimesMilliliter | TQuantityValueMilliAbsorbanceUnitTimesSecond | TQuantityValueMillivoltTimesSecond | TQuantityValueNanoCoulombTimesSecond | TQuantityValuePicoAmpereTimesSecond | None = (
        None
    )
    relative_peak_area: TQuantityValuePercent | None = None
    retention_time: TQuantityValueSecondTime | None = None
    retention_volume: TQuantityValueMilliliter | None = None
    peak_start: TQuantityValueMilliliter | TQuantityValueSecondTime | None = None
    peak_selectivity__chromatography_: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution: TQuantityValueUnitless | None = None
    field_index: int | None = None
    chromatographic_peak_resolution_using_baseline_peak_widths: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_peak_width_at_half_height__JP14_: TQuantityValueUnitless | None = (
        None
    )
    peak_width_at_4_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_13_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_32_4___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_60_7___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_half_height: TQuantityValueSecondTime | None = None
    peak_width_at_5___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_baseline: TQuantityValueSecondTime | None = None
    peak_width_at_inflection: TQuantityValueSecondTime | None = None
    peak_width_at_10___of_height: TQuantityValueSecondTime | None = None
    peak_width: TQuantityValueSecondTime | None = None
    statistical_skew__chromatography_: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_5___height: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_10___height: TQuantityValueUnitless | None = None
    asymmetry_factor_squared_measured_at_10___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_squared_measured_at_4_4___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_measured_at_4_4___height: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_baseline: TQuantityValueUnitless | None = None
    chromatographic_peak_asymmetry_factor: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution_using_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    chromatographic_peak_resolution_using_statistical_moments: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates__chromatography_: TQuantityValueUnitless | None = None
    number_of_theoretical_plates_measured_at_60_7___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_32_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_13_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_4_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_tangent_method: TQuantityValueUnitless | None = None
    chromatogram_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class PeakList:
    peak: list[Peak] | None = None


@dataclass(kw_only=True)
class ChromatographyColumnDocument:
    chromatography_column_part_number: TStringValue | None = None
    chromatography_column_serial_number: TStringValue | None = None
    chromatography_column_length: TQuantityValueCentimeter | None = None
    column_inner_diameter: TQuantityValueMillimeter | None = None
    chromatography_column_chemistry_type: TStringValue | None = None
    chromatography_column_particle_size: TQuantityValueMicrometer | None = None
    product_manufacturer: TStringValue | None = None


@dataclass(kw_only=True)
class InjectionDocument:
    autosampler_injection_volume_setting__chromatography_: TQuantityValueCubicMillimeter
    injection_identifier: TStringValue
    injection_time: TDateTimeValue


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem]


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    data_processing_document: dict[str, Any] | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None
    peak_list: PeakList | None = None
    derived_column_pressure_data_cube: DerivedColumnPressureDataCube | None = None
    chromatogram_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValue
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class ChromatogramDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class SystemFlowRateDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class SampleFlowRateDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class TemperatureProfileDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class SolventConcentrationDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class PreColumnPressureDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class PostColumnPressureDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class SamplePressureDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class SystemPressureDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class DerivedColumnPressureDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class MeasurementDocument:
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    chromatography_column_document: ChromatographyColumnDocument
    measurement_time: TDateTimeStampValue | None = None
    measurement_identifier: TStringValue | None = None
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    injection_document: InjectionDocument | None = None
    chromatogram_data_cube: ChromatogramDataCube | TDatacube | None = None
    three_dimensional_ultraviolet_spectrum_data_cube: TDatacube | None = None
    electropherogram_data_cube: TDatacube | None = None
    fluorescence_emission_profile_data_cube: TDatacube | None = None
    mass_chromatogram_data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocument]
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class LiquidChromatographyDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: TStringValue | None = None
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class LiquidChromatographyAggregateDocument:
    liquid_chromatography_document: list[LiquidChromatographyDocumentItem]
    data_system_document: DataSystemDocument | None = None
    device_system_document: DeviceSystemDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: AdmCoreREC202309ManifestSchema | str
    liquid_chromatography_aggregate_document: LiquidChromatographyAggregateDocument | None = (
        None
    )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.qpcr import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_calculated_documents import (
    iter_calculated_data_documents,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_data_creator import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_reader import (
    AppBioQuantStudioReader,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    create_amplification_data,
    create_multicomponent_data,
    Header,
    MeltCurveRawData,
    Result,
    Well,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "AppBio QuantStudio RT-PCR"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = AppBioQuantStudioReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = AppBioQuantStudioReader.create(named_file_contents)

        # Data sections must be read in order from the file.
        header = Header.create(reader.header)
        wells = Well.create(reader, header.experiment_type)
        amp_data = create_amplification_data(reader)
        multi_data = create_multicomponent_data(reader)
        results_data, results_metadata = Result.create(reader, header.experiment_type)
        melt_data = MeltCurveRawData.create(reader)

        calculated_data_documents = iter_calculated_data_documents(
            [well_item for well in wells for well_item in well.items],
            header.experiment_type,
            results_metadata.reference_sample_description,
            results_metadata.reference_dna_description,
        )

        return Data(
            metadata=create_metadata(header, named_file_contents.original_file_path),
            measurement_groups=create_measurement_groups(
                header, wells, amp_data, multi_data, results_data, melt_data
            ),
            calculated_data=create_calculated_data(
                calculated_data_documents, results_metadata
            ),
        )



"""Reader file for ThermoFisher Qubit 4 Adapter"""

import numpy as np
import pandas as pd

from allotropy.constants import DEFAULT_ENCODING
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv, read_excel, set_columns


class ThermoFisherQubit4Reader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"

    """
    A reader class for parsing ThermoFisher Qubit 4 data files into pandas DataFrames.

    Methods:
    read(named_file_contents: NamedFileContents) -> pd.DataFrame:
        Reads the content of the provided named file and returns it as a pandas DataFrame.
    """

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        """
        Reads the content of the provided named file and returns it as a pandas DataFrame.
        Changes the "Units" column to "Units_{previous_column}" for better understanding

        Parameters:
        named_file_contents (NamedFileContents): The named file contents to read.
            It includes the original file name, file contents, and encoding.

        Returns:
        pd.DataFrame: The content of the file as a pandas DataFrame.

        Raises:
        AllotropeConversionError: If the file format is not supported.
        """
        if named_file_contents.extension == "xlsx":
            df = read_excel(named_file_contents.contents)
        else:
            df = read_csv(
                named_file_contents.contents, index_col=False, encoding=DEFAULT_ENCODING
            )

        columns = df.columns.tolist()
        new_columns = [
            f"Units_{columns[i - 1]}" if "Units" in col else col
            for i, col in enumerate(columns)
        ]
        set_columns(df, new_columns)
        df = df.replace(np.nan, None)
        return df



# generated by datamodel-codegen:
#   filename:  liquid-chromatography.schema.json
#   timestamp: 2024-08-13T06:10:42+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCentimeter,
    TQuantityValueCubicMillimeter,
    TQuantityValueHertz,
    TQuantityValueMicrometer,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValuePercent,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacube,
    TDateTimeValue,
    TQuantityValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    pump_model_number: TStringValue | None = None
    detector_model_number: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None


@dataclass(kw_only=True)
class ChromatographyColumnDocument:
    chromatography_column_particle_size: TQuantityValueMicrometer | None = None
    chromatography_column_chemistry_type: TStringValue | None = None
    chromatography_column_serial_number: TStringValue | None = None
    column_inner_diameter: TQuantityValueMillimeter | None = None
    product_manufacturer: TStringValue | None = None
    chromatography_column_length: TQuantityValueCentimeter | None = None
    chromatography_column_part_number: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    model_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    detection_type: TStringValue | None = None
    electronic_absorbance_wavelength_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_reference_bandwidth_setting: TQuantityValueNanometer | None = (
        None
    )
    electronic_absorbance_reference_wavelength_setting: TQuantityValueNanometer | None = (
        None
    )
    detector_offset_setting: TQuantityValue | None = None
    detector_sampling_rate_setting: TQuantityValueHertz | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    description: Any | None = None
    written_name: TStringValue | None = None


@dataclass(kw_only=True)
class InjectionDocument:
    autosampler_injection_volume_setting__chromatography_: TQuantityValueCubicMillimeter
    injection_identifier: TStringValue
    injection_time: TDateTimeValue


@dataclass(kw_only=True)
class Peak:
    retention_time: TQuantityValueSecondTime
    peak_end: TQuantityValueSecondTime | None = None
    identifier: TStringValue | None = None
    relative_peak_height: TQuantityValuePercent | None = None
    written_name: TStringValue | None = None
    peak_height: TQuantityValue | None = None
    capacity_factor__chromatography_: TQuantityValueUnitless | None = None
    peak_area: TQuantityValue | None = None
    relative_peak_area: TQuantityValuePercent | None = None
    peak_start: TQuantityValueSecondTime | None = None
    peak_selectivity__chromatography_: TQuantityValueUnitless | None = None
    peak_width_at_4_4___of_height: TQuantityValueSecondTime | None = None
    field_index: int | None = None
    peak_width_at_13_4___of_height: TQuantityValueSecondTime | None = None
    statistical_skew__chromatography_: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_5___height: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_10___height: TQuantityValueUnitless | None = None
    asymmetry_factor_squared_measured_at_10___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_squared_measured_at_4_4___height: TQuantityValueUnitless | None = (
        None
    )
    asymmetry_factor_measured_at_4_4___height: TQuantityValueUnitless | None = None
    chromatographic_peak_asymmetry_factor: TQuantityValueUnitless | None = None
    asymmetry_factor_measured_at_baseline: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution: TQuantityValueUnitless | None = None
    chromatographic_peak_resolution_using_baseline_peak_widths: TQuantityValueUnitless | None = (
        None
    )
    peak_width_at_32_4___of_height: TQuantityValueSecondTime | None = None
    chromatographic_peak_resolution_using_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    chromatographic_peak_resolution_using_statistical_moments: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates__chromatography_: TQuantityValueUnitless | None = None
    number_of_theoretical_plates_measured_at_60_7___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_32_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_13_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_measured_at_4_4___of_peak_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_tangent_method: TQuantityValueUnitless | None = None
    number_of_theoretical_plates_by_peak_width_at_half_height: TQuantityValueUnitless | None = (
        None
    )
    number_of_theoretical_plates_by_peak_width_at_half_height__JP14_: TQuantityValueUnitless | None = (
        None
    )
    peak_width_at_60_7___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_half_height: TQuantityValueSecondTime | None = None
    peak_width_at_5___of_height: TQuantityValueSecondTime | None = None
    peak_width_at_baseline: TQuantityValueSecondTime | None = None
    peak_width_at_inflection: TQuantityValueSecondTime | None = None
    peak_width_at_10___of_height: TQuantityValueSecondTime | None = None
    peak_width: TQuantityValueSecondTime | None = None


@dataclass(kw_only=True)
class PeakList:
    peak: list[Peak] | None = None


@dataclass(kw_only=True)
class ProcessedDataDocument:
    peak_list: PeakList | None = None


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any
    data_cube: TDatacube | None = None


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    chromatography_column_document: ChromatographyColumnDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    injection_document: InjectionDocument
    detection_type: TStringValue
    chromatogram_data_cube: TDatacube
    measurement_identifier: TStringValue | None = None
    three_dimensional_ultraviolet_spectrum_data_cube: TDatacube | None = None
    processed_data_document: ProcessedDataDocument | None = None
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]


@dataclass(kw_only=True)
class LiquidChromatographyDocumentItem:
    analyst: TStringValue
    measurement_aggregate_document: MeasurementAggregateDocument
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class LiquidChromatographyAggregateDocument:
    device_system_document: DeviceSystemDocument
    liquid_chromatography_document: list[LiquidChromatographyDocumentItem]


@dataclass(kw_only=True)
class Model:
    manifest: str = "http://purl.allotrope.org/manifests/liquid-chromatography/REC/2023/03/liquid-chromatography.manifest"
    liquid_chromatography_aggregate_document: LiquidChromatographyAggregateDocument | None = (
        None
    )



from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

SOFTWARE_NAME = "“Bio-Plex Manager"
PRODUCT_MANUFACTURER = "Bio-Rad"
DEVICE_TYPE = "multi analyte profiling analyzer"
CONTAINER_TYPE = "well plate"

ERROR_MAPPING = {
    "1": "Low bead number",
    "2": "Aggregated beads",
    "3": "Classify efficiency",
    "4": "Region selection",
    "5": "Platform temperature",
}

SAMPLE_ROLE_TYPE_MAPPING = {
    "Blank": SampleRoleType.blank_role,
    "Control": SampleRoleType.control_sample_role,
    "Standard": SampleRoleType.standard_sample_role,
    "Unknown": SampleRoleType.unknown_sample_role,
}

EXPECTED_TAGS = [
    "Samples",
    "NativeDocumentLocation",
    "Description",
    "PlateDimensions",
    "Wells",
]



BRAND_NAME = "Qiacuity Digital PCR System"
PRODUCT_MANUFACTURER = "Qiagen"
SOFTWARE_NAME = "Qiacuity Software Suite"
DEVICE_TYPE = "dPCR"
DEVICE_IDENTIFIER = "Qiacuity dPCR"

SAMPLE_ROLE_TYPE_MAPPING = {
    "Sample": "Sample Role",
    "Control": "Control Sample Role",
    "Non Template Control": "Blank Role",
}



from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
import re
from typing import Any

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DATE_HEADER,
    DEFAULT_VERSION,
    HEADINGS_TO_PARSER_HEADINGS,
    MODEL_RE,
    XrVersion,
)
from allotropy.parsers.lines_reader import read_to_lines
from allotropy.parsers.utils.pandas import (
    assert_value_from_df,
    df_to_series,
    read_csv,
    read_excel,
    SeriesData,
)
from allotropy.parsers.utils.values import assert_not_none


@dataclass
class ViCellData:
    data: list[SeriesData]
    serial_number: str | None
    version: XrVersion


def create_reader_data(named_file_contents: NamedFileContents) -> ViCellData:
    reader: ViCellXRReader | ViCellXRTXTReader = (
        ViCellXRTXTReader(named_file_contents)
        if named_file_contents.extension == "txt"
        else ViCellXRReader(named_file_contents)
    )
    return ViCellData(reader.data, reader.serial_number, reader.version)


def _get_file_version(version_str: str | None) -> XrVersion:
    match = re.match(MODEL_RE, version_str or "", flags=re.IGNORECASE)
    if not match:
        return DEFAULT_VERSION
    try:
        version = assert_not_none(match).groupdict()["version"]
        # TODO: raise exception for unsupported versions
        version = ".".join(version.split(".")[0:2])
        return XrVersion(version)
    except AttributeError:
        return DEFAULT_VERSION


class ViCellXRReader:
    data: list[SeriesData]
    serial_number: str | None
    version: XrVersion

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        # calamine is faster for reading xlsx, but does not read xls. For xls, let pandas pick engine.
        self.engine = "calamine" if named_file_contents.extension == "xlsx" else None
        self.contents = named_file_contents.contents
        file_info = self._get_file_info()
        try:
            self.serial_number = (
                file_info.get(str, "serial", "").split(":", maxsplit=1)[1].strip()
            )
        except (IndexError, ValueError):
            self.serial_number = None
        self.version = _get_file_version(file_info.get(str, "model"))
        self.data = self._read_data()

    def _read_excel(self, **kwargs: Any) -> pd.DataFrame:
        return read_excel(self.contents, engine=self.engine, **kwargs)

    def _read_data(self) -> list[SeriesData]:
        header_row = 4
        if self.version == XrVersion._2_04:
            header_row = 3

        header = self._get_file_header(header_row)
        skiprows = header_row + 1

        file_data = self._read_excel(skiprows=skiprows, names=header)

        # rename the columns to match the existing parser that was based on xls(x) files
        file_data = file_data.rename(columns=HEADINGS_TO_PARSER_HEADINGS)

        # Do the datetime conversion and remove all rows that fail to pass as datetime
        # This fixes an issue where some files have a hidden invalid first row
        file_data[DATE_HEADER] = pd.to_datetime(
            assert_value_from_df(file_data, DATE_HEADER),
            format="%d %b %Y  %I:%M:%S %p",
            errors="coerce",
        )
        file_data = file_data.dropna(subset=DATE_HEADER)
        return [SeriesData(x) for _, x in file_data.iterrows()]

    def _get_file_header(self, header_row: int) -> list[str]:
        """Combine the two rows that forms the header."""
        header = (
            self._read_excel(
                nrows=2,
                skiprows=header_row,
                header=None,
            )
            .fillna("")
            .astype(str)
        )

        header_list: list[str] = header.agg(
            lambda x: " ".join(x).replace(" /ml", "/ml").strip()
        ).to_list()
        return header_list

    def _get_file_info(self) -> SeriesData:
        info: pd.Series[Any] = self._read_excel(
            nrows=3,
            header=None,
            usecols=[0],
        ).squeeze()
        info.index = pd.Index(["model", "filepath", "serial"])
        return SeriesData(info)


class ViCellXRTXTReader:
    data: list[SeriesData]
    serial_number: str | None
    version: XrVersion

    def __init__(self, contents: NamedFileContents) -> None:
        self.lines = read_to_lines(contents)
        self.data = self._read_data()
        self.serial_number = self.data[0].get(str, "Unit S/N")
        self.version = _get_file_version(str(self.data[0].series.name))

    def _read_data(self) -> list[SeriesData]:
        data_frame = read_csv(
            StringIO("\n".join(self.lines)), sep=" :", index_col=0, engine="python"
        )
        file_data = df_to_series(data_frame.astype(str).T)
        file_data = file_data.str.strip().replace("", None)
        file_data = file_data.rename(index=HEADINGS_TO_PARSER_HEADINGS)
        # Do the datetime conversion and remove all rows that fail to pass as datetime
        # This fixes an issue where some files have a hidden invalid first row
        file_data[DATE_HEADER] = pd.to_datetime(
            file_data[DATE_HEADER],
            format="%d %b %Y  %I:%M:%S %p",
            errors="coerce",
        )
        return [SeriesData(file_data)]



from typing import Any

import networkx as nx

from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)


def add_node(graph: nx.DiGraph, node_id: str, **attrs: Any) -> None:
    attrs["id"] = node_id[-3:]
    graph.add_node(node_id, **attrs)
    graph.nodes[node_id]["label"] = "\n".join(
        f"{key}: {attrs[key]}" for key in sorted(attrs)
    )


def construct_node(
    graph: nx.DiGraph, structure: dict[str, CalculatedDocument], calc_doc_id: str
) -> str:
    calc_doc = structure[calc_doc_id]
    add_node(
        graph,
        calc_doc.uuid,
        name=calc_doc.name,
        value=calc_doc.value,
        type="calculated document",
    )

    for data_source in calc_doc.data_sources:
        if sub_calc_doc := structure.get(data_source.reference.uuid):
            graph.add_edge(
                calc_doc.uuid,
                construct_node(graph, structure, sub_calc_doc.uuid),
            )
        else:
            node_id = f"{data_source.feature}_{data_source.reference.uuid}"
            add_node(
                graph,
                node_id,
                name=data_source.feature,
                value="unknown" if data_source.value is None else data_source.value,
                type="measurement",
            )
            graph.add_edge(calc_doc.uuid, node_id)

    return calc_doc.uuid


def construct_graph(
    calc_docs: dict[str, CalculatedDocument], target: str
) -> nx.DiGraph:
    graph = nx.DiGraph()

    for calc_doc_id, calc_doc in calc_docs.items():
        if calc_doc.name == target:
            construct_node(graph, calc_docs, calc_doc_id)

    return graph


def visualize_graph(
    calc_docs: dict[str, CalculatedDocument], calc_doc_name: str
) -> None:
    graph = construct_graph(calc_docs, calc_doc_name)
    nx.drawing.nx_agraph.write_dot(graph, "graph.dot")



from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from pathlib import Path
import re

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.tecan_magellan import constants
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none, try_float


def get_measurement_type(measurement_mode: str) -> MeasurementType:
    if measurement_mode.lower() == "absorbance":
        return MeasurementType.ULTRAVIOLET_ABSORBANCE

    msg = f"Only Absorbance measurements are supported as this time. Got {measurement_mode}."
    raise AllotropeConversionError(msg)


def parse_settings_lines(lines: list[str]) -> SeriesData:
    csv_stream = StringIO("\n".join(lines))
    raw_data = read_csv(
        csv_stream, header=None, sep=":", skipinitialspace=True, index_col=0
    )
    return df_to_series_data(raw_data.T)


@dataclass
class MeasurementSettings:
    measurement_mode: str
    measurement_type: MeasurementType
    wavelength_setting: float
    number_of_averages: float
    plate_identifier: str
    temperature: float

    @staticmethod
    def create(settings_lines: list[str], temperature: float) -> MeasurementSettings:
        settings = parse_settings_lines(settings_lines)
        raw_wavelength = settings[str, "Measurement wavelength"].split()[0]
        measurement_mode = settings[str, "Measurement mode"]

        return MeasurementSettings(
            measurement_mode=measurement_mode,
            measurement_type=get_measurement_type(measurement_mode),
            wavelength_setting=try_float(raw_wavelength, "Wavelength setting"),
            number_of_averages=settings[float, "Number of flashes"],
            plate_identifier=settings[str, "Plate definition file"].split(".")[0],
            temperature=temperature,
        )


@dataclass(frozen=True)
class MagellanMetadata:
    measurement_time: str
    analytical_method_identifier: str | None
    experimental_data_identifier: str | None
    analyst: str
    device_identifier: str
    equipment_serial_number: str
    measurements_settings: dict[str, MeasurementSettings]

    @classmethod
    def create(cls, metadata_lines: list[str]) -> MagellanMetadata:
        reader = LinesReader(metadata_lines)
        measurement_time = cls.parse_measurement_time(
            assert_not_none(reader.pop(), "Measurement Time")
        )
        analytical_method_identifier = cls._get_identifier_line(reader, r".*\.mth")
        experimental_data_identifier = cls._get_identifier_line(reader, r".*\.wsp")
        # skip wavelength lines
        reader.drop_until(r"^(?!\d+ ?nm).*$")
        analyst = assert_not_none(reader.pop(), "Analyst")
        device_identifier = assert_not_none(reader.pop(), "Device Identifier")
        raw_serial_number = assert_not_none(reader.pop(), "Equipment Serial Number")
        equipment_serial_number = raw_serial_number.split(": ")[1]

        return MagellanMetadata(
            measurement_time=measurement_time,
            analytical_method_identifier=analytical_method_identifier,
            experimental_data_identifier=experimental_data_identifier,
            analyst=analyst,
            device_identifier=device_identifier,
            equipment_serial_number=equipment_serial_number,
            measurements_settings=cls._get_measurements_settings(reader),
        )

    @classmethod
    def parse_measurement_time(cls, raw_datetime: str) -> str:
        if match := re.match(constants.MEASUREMENT_TIME_REGEX, raw_datetime):
            return f"{match.groups()[0]} {match.groups()[1]}"
        msg = "Unable to get measurement time from {raw_datetime}"
        raise AllotropeConversionError(msg)

    @classmethod
    def _get_identifier_line(cls, reader: LinesReader, match_pat: str) -> str | None:
        if lines := list(reader.pop_while(match_pat)):
            return lines[-1]
        return None

    @classmethod
    def _get_measurements_settings(
        cls, reader: LinesReader
    ) -> dict[str, MeasurementSettings]:
        settings_lines: list[list[str]] = []
        settings: dict[str, MeasurementSettings] = {}
        label_and_temps = []
        temp_line_re = r"Meas. temperature: (.+): ([\d\.]+) °C"

        while not re.match(temp_line_re, (first_line := assert_not_none(reader.get()))):
            if match := re.match(r"Measurement \d.", first_line):
                reader.pop()
            settings_lines.append(list(reader.pop_until_inclusive(r"Unit:.+")))

        while match := re.match(temp_line_re, assert_not_none(reader.pop())):
            label_and_temps.append(match.groups())

        if len(label_and_temps) != len(settings_lines):
            msg = "Number of temperatures does not match number of measurements."
            raise AllotropeConversionError(msg)

        for idx, lines in enumerate(settings_lines):
            label, temperature = label_and_temps[idx]
            settings[str(label)] = MeasurementSettings.create(
                lines, try_float(str(temperature), "Measurement temperature")
            )

        return settings


def create_metadata(data: MagellanMetadata, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        unc_path=file_path,
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        device_identifier=data.device_identifier,
        model_number=NOT_APPLICABLE,
        data_system_instance_id=NOT_APPLICABLE,
        software_name=constants.SOFTWARE_NAME,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        equipment_serial_number=data.equipment_serial_number,
    )


def create_measurement_groups(
    data: SeriesData, metadata: MagellanMetadata, well_count: float
) -> MeasurementGroup:
    measurements = []

    for measurement_label, settings in metadata.measurements_settings.items():
        errors = []
        if (measurement := data.get(float, measurement_label)) is None:
            measurement = NEGATIVE_ZERO
            errors = [
                ErrorDocument(
                    error=data[str, measurement_label],
                    error_feature=settings.measurement_mode,
                )
            ]

        location_identifier = data[str, "Well positions"]
        well_plate_identifier = data.get(str, "Plate", settings.plate_identifier)
        measurements.append(
            Measurement(
                type_=settings.measurement_type,
                device_type=constants.DEVICE_TYPE,
                identifier=random_uuid_str(),
                sample_identifier=f"{well_plate_identifier}_{location_identifier}",
                location_identifier=location_identifier,
                well_plate_identifier=well_plate_identifier,
                detection_type=settings.measurement_mode,
                compartment_temperature=settings.temperature,
                absorbance=measurement,
                number_of_averages=settings.number_of_averages,
                detector_wavelength_setting=settings.wavelength_setting,
                error_document=errors,
            )
        )

    return MeasurementGroup(
        measurements=measurements,
        plate_well_count=well_count,
        measurement_time=metadata.measurement_time,
        analyst=metadata.analyst,
        analytical_method_identifier=metadata.analytical_method_identifier,
        experimental_data_identifier=metadata.experimental_data_identifier,
    )



from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Data as MapperData,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.luminex_xponent.luminex_xponent_reader import (
    LuminexXponentReader,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_structure import (
    create_measurement_groups,
    create_metadata,
    Data,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class LuminexXponentParser(VendorParser[MapperData, Model]):
    DISPLAY_NAME = "Luminex xPONENT"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = LuminexXponentReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> MapperData:
        reader = LuminexXponentReader(named_file_contents)
        data = Data.create(reader)
        return MapperData(
            create_metadata(
                data.header, data.calibrations, named_file_contents.original_file_path
            ),
            create_measurement_groups(data.measurement_list.measurements, data.header),
        )



from collections.abc import Mapping
import importlib
from pathlib import Path, PurePosixPath, PureWindowsPath
import re
from typing import Any

ALLOTROPE_DIR: Path = Path(__file__).parent.parent
ALLOTROPY_DIR: Path = ALLOTROPE_DIR.parent
ROOT_DIR: Path = ALLOTROPE_DIR.parent.parent.parent
SCHEMA_DIR_PATH: Path = Path(ALLOTROPE_DIR, "schemas")
SHARED_SCHEMAS_PATH: Path = Path(SCHEMA_DIR_PATH, "shared")
SHARED_SCHEMAS_DEFINITIONS_PATH: Path = Path(SHARED_SCHEMAS_PATH, "definitions")
MODEL_DIR_PATH: Path = Path(ALLOTROPE_DIR, "models")
SHARED_MODELS_PATH: Path = Path(MODEL_DIR_PATH, "shared")
SHARED_MODELS_DEFINITIONS_PATH: Path = Path(SHARED_MODELS_PATH, "definitions")
GENERATED_SHARED_PATHS: list[Path] = [
    Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "units.json"),
    Path(SHARED_MODELS_DEFINITIONS_PATH, "units.py"),
    Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "custom.json"),
    Path(SHARED_MODELS_DEFINITIONS_PATH, "custom.py"),
]


def get_rel_schema_path(schema_path: Path) -> Path:
    try:
        return schema_path.relative_to(SCHEMA_DIR_PATH)
    except ValueError as err:
        if not Path(SCHEMA_DIR_PATH, schema_path).exists():
            msg = f"Invalid schema path: {schema_path}"
            raise ValueError(msg) from err
        return schema_path


def get_rel_model_path(model_path: Path) -> Path:
    try:
        return model_path.relative_to(MODEL_DIR_PATH)
    except ValueError as err:
        if not Path(MODEL_DIR_PATH, model_path).exists():
            msg = f"Invalid model path: {model_path}"
            raise ValueError(msg) from err
        return model_path


def get_full_schema_path(schema_path: Path) -> Path:
    if str(schema_path).startswith(str(SCHEMA_DIR_PATH)):
        return schema_path
    return Path(SCHEMA_DIR_PATH, schema_path)


def get_manifest_from_schema_path(schema_path: Path) -> str:
    rel_schema_path = get_rel_schema_path(schema_path)
    if not rel_schema_path.parts[0] == "adm" or not str(rel_schema_path).endswith(
        ".schema.json"
    ):
        msg = f"Invalid schema path: {rel_schema_path}"
        raise ValueError(msg)
    return f"http://purl.allotrope.org/manifests/{str(PurePosixPath(rel_schema_path))[4:-12]}.manifest"


def get_manifest_from_model_path(model_path: Path) -> str:
    return get_manifest_from_schema_path(get_schema_path_from_model_path(model_path))


def get_schema_path_from_manifest(manifest: str) -> Path:
    match = re.match(r"http://purl.allotrope.org/manifests/(.*)\.manifest", manifest)
    if not match:
        msg = f"No matching schema in repo for manifest: {manifest}"
        raise ValueError(msg)
    return Path(f"adm/{match.groups()[0]}.schema.json")


def get_schema_path_from_reference(reference: str) -> Path:
    ref_match = re.match(r"http://purl.allotrope.org/json-schemas/(.*)", reference)
    if not ref_match:
        msg = f"Could not parse reference: {reference}"
        raise ValueError(msg)
    return Path(f"{ref_match.groups()[0]}.json".replace(".embed", ""))


def get_model_path_from_schema_path(schema_path: Path) -> Path:
    rel_schema_path = PureWindowsPath(get_rel_schema_path(schema_path))
    schema_file = rel_schema_path.name
    model_file = schema_file.replace(".schema.json", ".py").replace("-", "_")
    model_path = Path(
        *[
            re.sub("^([0-9]+)$", r"_\1", part.lower().replace("-", "_"))
            for part in rel_schema_path.parent.parts
        ]
    )
    return Path(model_path, model_file)


def get_schema_path_from_model_path(model_path: Path) -> Path:
    rel_model_path = PureWindowsPath(get_rel_model_path(model_path))
    model_file = rel_model_path.name
    model_file = model_file.replace(".py", ".schema.json").replace("_", "-")
    model_path_parts = [
        re.sub("^_([0-9]+)$", r"\1", part).replace("_", "-")
        for part in rel_model_path.parent.parts
    ]
    model_path_parts[2] = model_path_parts[2].upper()
    model_path = Path(*model_path_parts)
    return Path(model_path, model_file)


def get_import_path_from_path(model_path: Path) -> str:
    # NOTE: PureWindowsPath handles both Windows and linux paths.
    return (
        f"allotropy.allotrope.models.{'.'.join(PureWindowsPath(model_path).parts)[:-3]}"
    )


def get_model_class_from_schema(asm: Mapping[str, Any]) -> Any:
    schema_path = get_schema_path_from_manifest(asm["$asm.manifest"])
    model_path = get_model_path_from_schema_path(Path(schema_path))
    import_path = get_import_path_from_path(model_path)
    # NOTE: it is safe to assume that every schema module has Model, as we generate this code.
    return importlib.import_module(import_path).Model



from decimal import Decimal
from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.chemometec_nc_view import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        file_name=path.name,
        unc_path=file_path,
        software_name=constants.SOFTWARE_NAME,
        device_type=constants.DEVICE_TYPE,
        equipment_serial_number=data[str, "INSTRUMENT"].split(":")[-1].strip(),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        detection_type=constants.DETECTION_TYPE,
        model_number=NOT_APPLICABLE,
    )


def create_measurement_groups(data: SeriesData) -> MeasurementGroup:
    # Cell counts are measured in cells/mL, but reported in millions of cells/mL
    viable_cell_density = float(
        Decimal(_format_unit(data[str, "LIVE (cells/ml)"])) / Decimal("1000000")
    )
    total_cell_density_val = data.get(str, "TOTAL (cells/ml)")
    if total_cell_density_val:
        total_cell_density = float(
            Decimal(_format_unit(total_cell_density_val)) / Decimal("1000000")
        )
    dead_cell_density_val = data.get(str, "DEAD (cells/ml)")
    if dead_cell_density_val:
        dead_cell_density = float(
            Decimal(_format_unit(dead_cell_density_val)) / Decimal("1000000")
        )

    return MeasurementGroup(
        analyst=data[str, "OPERATOR"],
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                processed_data_identifier=random_uuid_str(),
                timestamp=_format_timestamp(data[str, "NAME"]),
                sample_identifier=data[str, "SAMPLE ID"],
                viability=data[float, "VIABILITY (%)"],
                total_cell_density=total_cell_density,
                viable_cell_density=viable_cell_density,
                dead_cell_density=dead_cell_density,
                average_total_cell_diameter=try_float_or_none(
                    data.get(str, "DIAMETER (μm)")
                ),
                cell_aggregation_percentage=try_float_or_none(
                    data.get(str, "AGGREGATES (%)")
                ),
                debris_index=data.get(float, "DEBRIS INDEX"),
                cell_density_dilution_factor=data.get(float, "DILUTION FACTOR"),
            )
        ],
    )


def _format_unit(unit: str) -> float:
    return float("".join(unit.split()))


def _format_timestamp(timestamp: str) -> str:
    return timestamp.split("-")[0]



from typing import ClassVar

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
    WellList,
)


class MeltCurveCreator(Creator):
    EXPECTED_SHEETS: ClassVar[list[str]] = [
        "Melt Curve Raw",
        "Melt Curve Result",
    ]

    @classmethod
    def check_plugin_name(cls, _: str | None) -> bool:
        return True

    @classmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        header = Header.create(reader.header)
        wells = WellList.create(reader, header)
        return Data(
            header,
            wells,
            experiment_type=ExperimentType.melt_curve_qPCR_experiment,
            calculated_documents=[],
            reference_target=None,
            reference_sample=None,
        )



NOT_APPLICABLE = "N/A"
# Used to fill an error value in a required ASM field.
NEGATIVE_ZERO = -0.0
# Used to fill in timestamp for required ASM fields when not available in source
DEFAULT_EPOCH_TIMESTAMP = "1970-01-01"



#!/usr/bin/env python3
from pathlib import Path

import click

from allotropy.allotrope.schema_parser.path_util import (
    get_import_path_from_path,
    get_manifest_from_schema_path,
    get_model_path_from_schema_path,
    SCHEMA_DIR_PATH,
)
from allotropy.allotrope.schema_parser.schema_model import snake_to_upper_camel

ALLOTROPY_DIR = Path(Path(__file__).parent.parent, "src/allotropy")
TEMPLATE_DIR = Path(Path(__file__).parent, "templates")


def write_template_files(
    output_dir: Path,
    template_replacements: dict[str, str],
    template_files: dict[str, str],
) -> None:
    for template_file, output_file in template_files.items():
        with open(Path(TEMPLATE_DIR, template_file)) as f:
            template = f.read()

        with open(Path(output_dir, output_file), "w") as f:
            for pattern, value in template_replacements.items():
                template = template.replace(f"${pattern}$", value)

            f.write(template)


def add_to_parser_factory(parser_name: str, enum_name: str, class_name: str) -> None:
    parser_factory_file = Path(ALLOTROPY_DIR, "parser_factory.py")
    with open(parser_factory_file) as f:
        contents = f.readlines()

    line_and_condition = {
        f'    {enum_name} = "{enum_name}"\n': "class Vendor",
        f"    Vendor.{enum_name}: {class_name},\n": "_VENDOR_TO_PARSER",
        f"from allotropy.parsers.{parser_name}.{parser_name}_parser import {class_name}\n": "from allotropy.parsers",
    }
    in_condition: dict[str, bool] = {}

    with open(parser_factory_file, "w") as f:
        for line in contents:
            for new_line, condition in line_and_condition.items():
                if new_line not in in_condition and line.startswith(condition):
                    in_condition[new_line] = True
                elif in_condition.get(new_line) and new_line <= line:
                    if new_line != line:
                        f.write(new_line)
                    in_condition[new_line] = False

            f.write(line)


def create_parser(
    name: str, schema_path: Path, display_name: str | None = None
) -> None:
    name = name.replace(" ", "_").replace("-", "_").lower()
    enum_name = name.upper()
    display_name = display_name or name.replace("_", " ").title()
    class_name_prefix = f"{snake_to_upper_camel(name)}"

    model_path = get_model_path_from_schema_path(schema_path)
    import_path = get_import_path_from_path(model_path)
    manifest = get_manifest_from_schema_path(schema_path)

    parser_dir = Path(ALLOTROPY_DIR, "parsers", name)
    tests_dir = Path(ALLOTROPY_DIR.parent.parent, "tests/parsers", name)
    if parser_dir.exists():
        msg = f"Parser {name} already exists!"
        raise ValueError(msg)

    parser_dir.mkdir()
    Path(parser_dir, "__init__.py").touch()
    tests_dir.mkdir()
    Path(tests_dir, "__init__.py").touch()
    Path(tests_dir, "testdata").mkdir()

    template_replacements = {
        "PARSER_NAME": name,
        "CLASS_NAME_PREFIX": class_name_prefix,
        "DISPLAY_NAME": display_name,
        "ENUM_NAME": enum_name,
        "MODEL_IMPORT_PATH": import_path,
        "MAPPER_IMPORT_PATH": import_path.replace("models", "schema_mappers"),
        "MANIFEST": manifest,
    }
    write_template_files(
        parser_dir,
        template_replacements,
        {
            "parser_template": f"{name}_parser.py",
            "structure_template": f"{name}_structure.py",
            "reader_template": f"{name}_reader.py",
            "constants_template": "constants.py",
        },
    )
    write_template_files(
        tests_dir,
        template_replacements,
        {
            "test_template": "to_allotrope_test.py",
        },
    )
    add_to_parser_factory(name, enum_name, f"{class_name_prefix}Parser")


@click.command()
@click.argument("name")
@click.argument("schema_regex")
@click.option(
    "--display_name", help="Display name for parser, defaults to title case of NAME"
)
def _create_parser(
    name: str, schema_regex: str, display_name: str | None = None
) -> None:
    """Create parser with name NAME for schema found with SCHEMA_REGEX."""
    if "parser" in name.lower():
        msg = "Cannot include 'parser' in name"
        raise ValueError(msg)

    schema_paths = list(SCHEMA_DIR_PATH.rglob(f"{schema_regex}*.json"))
    if not schema_paths:
        msg = f"Could not find schema with schema regex: '{schema_regex}'"
        raise ValueError(msg)
    if len(schema_paths) > 1:
        msg = f"Found more than one schema with schema regex, please narrow down: {[str(p) for p in schema_paths]}"
        raise ValueError(msg)

    create_parser(name, schema_paths[0], display_name)

    from allotropy.parser_factory import update_supported_instruments

    update_supported_instruments()


if __name__ == "__main__":
    _create_parser()
